"""Test the events framework."""

from pytest import raises
from carehome import Database, methods
from carehome.exc import NoSuchEventError

db = Database()


class InitError(Exception):
    pass


class DestroyError(Exception):
    pass


# Make these exceptions accessible to Object methods:
methods.InitError = InitError
methods.DestroyError = DestroyError


def test_valid_event():
    o = db.create_object()
    o.add_method(
        'on_event', 'return (self, args, kwargs)', args='self, *args, **kwargs'
    )
    self, args, kwargs = o.do_event('on_event', 1, 2, 3, hello='world')
    assert self is o
    assert args == (1, 2, 3)
    assert kwargs == {'hello': 'world'}


def test_invalid_event():
    o = db.create_object()
    with raises(NoSuchEventError):
        o.do_event('test_event')


def test_try_event_valid():
    o = db.create_object()
    o.add_method(
        'on_event', 'return (self, args, kwargs)', args='self, *args, **kwargs'
    )
    self, args, kwargs = o.try_event('on_event', 1, 2, 3, hello='world')
    assert self is o
    assert args == (1, 2, 3)
    assert kwargs == {'hello': 'world'}


def test_try_event_invalid():
    o = db.create_object()
    assert o.try_event('test_event') is None


def test_on_init():
    p = db.create_object()
    p.add_method('on_init', 'raise InitError()')
    with raises(InitError):
        p.on_init()
    assert 'on_init' in p.methods
    with raises(InitError):
        db.create_object(p)


def test_on_destroy():
    o = db.create_object()
    o.add_method('on_destroy', 'raise DestroyError()')
    with raises(DestroyError):
        db.destroy_object(o)
    assert o.id in db.objects


def test_on_exit():
    loc = db.create_object()
    loc.add_method('on_exit', 'self.last_left = thing', args='self, thing')
    thing = db.create_object()
    thing.location = loc
    assert not loc.properties
    thing.location = None
    assert loc.last_left is thing
    other = db.create_object()
    other.location = loc
    assert loc.last_left is thing
    other.location = None
    assert loc.last_left is other


def test_on_enter():
    loc = db.create_object()
    loc.add_method('on_enter', 'self.last_entered = thing', args='self, thing')
    assert not loc.properties
    thing = db.create_object()
    thing.location = loc
    assert loc.last_entered is thing
    assert loc.contents == [thing]
    other = db.create_object()
    other.location = loc
    assert loc.last_entered is other
    assert loc.contents == [thing, other]



"""Test the Method class."""

import re
from types import FunctionType
from carehome import Method, Database

db = Database()
inserted_global = object()


def test_init():
    m = Method(db, 'test', 'Test method.', '', [], 'print("Hello world.")')
    assert isinstance(m.func, FunctionType)
    assert m.name == 'test'
    f = m.func
    assert f.__name__ == 'test'
    assert f.__doc__ == m.description


def test_run():
    m = Method(db, 'test', 'Test function.', '', [], 'return 12345')
    assert m.func() == 12345


def test_args():
    m = Method(db, 'test', 'Test function.', 'a, b=5', [], 'return (a, b)')
    assert m.func(1) == (1, 5)
    assert m.func(4, b=10) == (4, 10)


def test_imports():
    m = Method(db, 'test', 'Test re.', '', ['import re'], 'return re')
    assert m.func() is re


def test_method_globals():
    db = Database(method_globals=dict(g=inserted_global))
    o = db.create_object()
    o.add_method('get_g', 'return g')
    assert o.get_g() is inserted_global



"""Provides exception classes."""


class CarehomeError(Exception):
    """All other errors should inherit from this class."""


class InheritanceError(CarehomeError):
    """An error in the inheritance hierarchy."""


class DuplicateParentError(InheritanceError):
    """This object already has this parent."""


class DuplicateChildError(InheritanceError):
    """This object already has this child."""


class ParentIsChildError(InheritanceError):
    """This parent is already a child of this object."""


class DatabaseError(CarehomeError):
    """There is a problem in the database."""


class NoSuchEventError(CarehomeError):
    """No such event."""



"""Provides the Method class."""

from attr import attrs, attrib, Factory


@attrs
class Method:
    """An Object method."""

    database = attrib()
    name = attrib()
    description = attrib()
    args = attrib()
    imports = attrib()
    code = attrib()
    func = attrib(default=Factory(type(None)), init=False)

    def __attrs_post_init__(self):
        g = globals().copy()
        g.update(**self.database.method_globals)
        g['objects'] = self.database.objects
        code = '\n'.join(self.imports)
        code += '\ndef %s(%s):\n    """%s"""' % (
            self.name, self.args, self.description
        )
        for line in self.code.splitlines():
            code += '\n    '
            code += line
        source = compile(code, f'<Method {self.name}>', 'exec')
        eval(source, g)
        self.func = g[self.name]
